/* $Workfile: A66EchoCancel.c $       $Revision: 14 $      $Modtime: 12/14/05 6:04p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    A66EchoCancel.c

    A basic example console application demonstrating the use of OTX Driver
    with the Arni-6x6-PCI or the Arni-6x6-PCI board.

    The sample application configures 1 FXO POTS Line Interface.

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    Note that the Echo Cancel device takes to input connections. We use
    the ConnectionPoint parameter of the OtxDrvConnectLogicalDevice() call
    to determine which of the two input connections that we are connecting
    to. The two connections are defined as:
    
      Connection Point 0: The Near End input (signal with hybrid echo)
      Connection Point 1: The Far End input (reference/prompt signal)

    Note also that only logical devices have connections points. Physical
    device do not.

    We use the Burst device to record the output from the Echo cancel device 
    (after the hybrid echo has been removed/cancelled). The recorded samples are 
    stored in u-law format.

    We also use the Burst device to play the reference/prompt signal out to 
    FXO device. This is the signal that should be removed/cancelled by the Echo
    Cancel device. This demo will pick up the prompt signal from a file and play
    it continously (looped). The samples in the file are stored in u-law format.

    Also while we are playing this reference/prompt file to the FXO device, we are
    also delaying it 8 ms, and then sending the same file to the Echo Cancel
    device. The delay is important for the Echo Cancel device to work properly. The
    8ms delay is emperically chosen to work well with the Arni-6x6-PCI board. However,
    for certain application it might be necessary to tune the delay for optimum 
    performance of the Echo Cancel device. 

    We use a different timeslot of the burst highway for the delayed reference/prompt
    file signal.

    The following connections are made:
    
    pipe = hDaaRx, which contains
    received samples from the DAA device (signal with hybrid echo).
    
      INPUT pipe end:  PotsLi device     (ConnPoint = -1) DSPHwyTs=0 
      OUTPUT pipe end: EchoCancel device (ConnPoint = 0)  DSPHwyTs=1

    pipe = hDaaTx, which contains
    the reference/prompt samples to be output to the DAA device. The samples are
    provided by the burst device.
    
      INPUT pipe end:  Burst device                       BurstHwyTs=0
      OUTPUT pipe end: PotsLi device     (ConnPoint = -1) DSPHwyTs=0

    pipe = hBurstTx, which contains
    the delayed reference/prompt samples to be the Near Input to the Echo
    Cancel device. The samples are provided by the burst device. The application
    will buffer (delay) these samples by 8ms. Other than the delay, the data
    that travels through this pipe is the same as for the hDaaTx pipe.
    
      INPUT pipe end:  Burst device                       BurstHwyTs=1
      OUTPUT pipe end: EchoCancel device (ConnPoint = 1)  DSPHwyTs=2

    pipe = hBurstRx, which contains
    the samples from the Echo Cancel device (after the hybrid echo has been
    remove/cancelled).
    
      INPUT pipe end:  EchoCancel device (ConnPoint = -1) DSPHwyTs=10
      OUTPUT pipe end: Burst device                       BurstHwyTs=0

    Please note that all of the local highways on the board are fully duplex.
    Therefore we can use TS0 for the same highway for two different pipes
    (one for each direction).

    The main command keys are listed in the following:
     - Press 'h' or '?' for help. 'q' to quit.
     

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxBrd.h"
#include "OtxCst.h"
#include "OtxDsp.h"
#include "OtxBurst.h"
#include "OtxBurstD.h"



/*******************************************************************************
  Macros, Types
*******************************************************************************/

//#define DEMO_EC_DEBUG   // Define to use the burst device to provide both inputs streams to the Echo Cancel device (i.e. exclude the DAA device)

#define DEMO_DSP_MAX 1
#define DEMO_LI_MAX 1

#define DEMO_TASK_REF           1L                      // Task Reference for a non-blocking call
#define DEMO_DEFAULT_LOG_FILE   "A66EchoCancel.log"     // Name of the LOG file

#ifdef DEMO_EC_DEBUG
#define DEMO_EC_FE_INPUT_FILE   "ref_sp.snd"    
#define DEMO_EC_ECHO_INPUT_FILE "echo_sp.snd"    
#else
//#define DEMO_EC_FE_INPUT_FILE   "ref_sp2.snd"         // U-law samples of far end input to the Echo Cancel device (i.e. prompt file to be cancelled by the Echo Cancel device). The program plays this file from disk.
#define DEMO_EC_FE_INPUT_FILE   "PromptFileUlaw.snd"    // U-law samples of far end input to the Echo Cancel device (i.e. prompt file to be cancelled by the Echo Cancel device). The program plays this file from disk.
#endif

#define DEMO_EC_NE_REC_FILE     "EcNeUlaw.snd"          // Recorded U-law samples of near end input to the Echo Cancel device. This file is recorded from the DAA samples (audio) provided on the Li.
#define DEMO_EC_FE_REC_FILE     "EcFeUlaw.snd"          // Recorded U-law samples of far end input to the Echo Cancel device. This file is recorded from the DAA samples (audio) provided on the Li.
#define DEMO_EC_OUTPUT_REC_FILE "EcOutputUlaw.snd"      // Recorded U-law samples of output from the Echo Cancel device, where the far end input samples are cancelled. This file is recorded by this program.
#define DEMO_DEFAULT_BOARD_NO   0
#define DEMO_DEFAULT_FXO_NO     2

#define DEMO_BURST_BUF_SIZE     0x8000    // 32k

#define DEMO_U_LAW_SILENCE   0xFF    // Sending this byte produces silence in u-law mode
#define DEMO_A_LAW_SILENCE   0xD5    // Sending this byte produces silence in a-law mode


typedef struct DemoEnvT {
  char           *pszInputStr;
  OTX_BOOL        bExitOnError;
  char           *logFileName;
  char           *outFileName;
  int             boardNo;
  int             fxoNo;
  OtxPhysicalDeviceTypeE  nBoardType;
  OtxDataFormatE  law;
  OtxCstPotsCfgSetE  nSetId;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static char DemoDtmfParamToChar(int nParam);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static OTX_INT32 DemoProcessUserCommand(char commandChar);
static int DemoExit(int nExitCode);

/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue        = OTX_INVALID_HANDLE_VALUE;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
FILE       *g_fStreamFile = NULL;
OTX_BOOL    g_bWatchDogThreadActive = OTX_FALSE;

static OsaMutexS g_demoLogfCritSec;

struct {  // Handles to Physical Devices
  OTX_HANDLE hBoardController0;
  OTX_HANDLE hSlicLi[6];
  OTX_HANDLE hDsp[6];
  OTX_HANDLE hBurst;
  OTX_HANDLE hTss;
} g_hPhysDevs;


// direct Burst DMA send
struct sBurstSender {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  OTX_UINT8   sqTxBuf[DEMO_BURST_BUF_SIZE];
  OTX_UINT8   sqBufCurr[DEMO_BURST_BUF_SIZE / 32]; // For file reads
  OTX_UINT8   sqBufPrev[DEMO_BURST_BUF_SIZE / 32]; // For file reads (the previous buffer)
  FILE        *pTxFile;  // EC Far End Input (from DAA)
#ifdef DEMO_EC_DEBUG
  FILE        *pTxFile2;  // Echo Input (from file instead of from DAA)
#endif
  OTX_UINT32  cbTxBuf;
  OTX_BOOL    bFileLoop;
};

// direct Burst DMA receive
struct sBurstReceiver {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  FILE        *pRxFile;   // Recorded EC output
  FILE        *pRxFile2;  // Recorded EC Near End Input
  FILE        *pRxFile3;  // Recorded EC Far End Input
  OTX_INT32   nTsToRecord;  // Default is -1 (All TS)
};


// Pots Li device
struct sPotsLi {
  OTX_BOOL    bOffHook;
};

struct {  // Handles to Logical Devices
  OTX_HANDLE hPotsLi[DEMO_LI_MAX];
  OTX_HANDLE hEchoCancel[DEMO_LI_MAX];
  OTX_HANDLE hUserApp[DEMO_LI_MAX];
} g_hLogicDevs;

struct {
  struct sBurstSender mBurstSender;        
  struct sBurstReceiver mBurstReceiver;    
} g_sBurstDevs;


struct {
  struct sPotsLi mPotsLi;        
} g_sPotsDevs[DEMO_LI_MAX];

struct {  // Handles to Pipes
  OTX_HANDLE hDaaRx[DEMO_LI_MAX];
  OTX_HANDLE hDaaTx[DEMO_LI_MAX];
  OTX_HANDLE hBurstRx[DEMO_LI_MAX];
  OTX_HANDLE hBurstTx[DEMO_LI_MAX];
} g_hPipes;

/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************
  @func
    Create the event queue to be used by the driver to inform the application
    about events, such as line status changes, or detected DMTF digits.
*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

}



void DemoOpenPhysicalDevices(void)
/*******************************************************************************
  @func
    Opens the Physical Devices on the Arni-16-PCI board

  @comm
    Calls function <f OtxDrvOpenPhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Get Handle to the Physical ARNI Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_ARNI_6X6_PCI_PLUS:
    DemoLogf("Opening Arni-6x6-PCI-Plus board\n");
    break;
  case OTX_DEVICE_ARNI_6_FXO_PCI_PLUS:
    DemoLogf("Opening Arni-6-FXO-PCI-Plus board\n");
    break;
  case OTX_DEVICE_ARNI_6X3_PCI_PLUS:
    DemoLogf("Opening Arni-6x3-PCI-Plus board\n");
    break;
  case OTX_DEVICE_ARNI_6X6_PCI:
    DemoLogf("Opening Arni-6x6-PCI board\n");
    break;
  case OTX_DEVICE_ARNI_6_FXS_PCI_PLUS:
    DemoLogf("Opening Arni-6-FXS-PCI-Plus board\n");
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", 
             g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hBoardController0); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Check if the watchdog timer has triggered before we continue
  {
    OTX_UINT32 readVal;
    result = OtxDrvReadReg32(g_hPhysDevs.hBoardController0, 0x04, &readVal); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    if (readVal == 0) {
      DemoLogf("The watchdog timer has triggered. A reboot of the PC is necessary.\n");
      DemoExitQuestion();
    }
  }


  // Get a handle to the Time-Space Switch (TSS) on the board (just for debugging purposes)
  DemoLogf("Open OTX_DEVICE_TSS\n");
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_TSS, 0, 0, g_hEventQueue, &g_hPhysDevs.hTss); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  for (i = 0; i < DEMO_DSP_MAX; i++) {
    DemoLogf("Open OTX_DEVICE_DSP #%d\n", g_demoEnv.fxoNo);
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_DSP, g_demoEnv.fxoNo, g_demoEnv.fxoNo, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }

  for (i = 0; i < DEMO_DSP_MAX; i++) {
    if (g_demoEnv.outFileName) {
      DemoLogf("Loading program file %s to DSP#%d \n", g_demoEnv.outFileName, g_demoEnv.fxoNo);
      result = OtxDspRunProgram(g_hPhysDevs.hDsp[i], g_demoEnv.outFileName, "_c_int00"); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    } else {
      DemoLogf("Loading OTX_DSP_SPM_OTXCST2 to DSP#%d \n", g_demoEnv.fxoNo);
      result = OtxDspRunSpm(g_hPhysDevs.hDsp[i], OTX_DSP_SPM_OTXCST2); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    }
  }

  // Get a handle to the physical Burst device.
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hBoardController0, OTX_DEVICE_BURST, 0, 0, g_hEventQueue, &g_hPhysDevs.hBurst); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // set device buffer event trigger levels
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,  DEMO_BURST_BUF_SIZE); DemoCheckRc(g_hPhysDevs.hBurst, result);
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_TRIGGER,    DEMO_BURST_BUF_SIZE); DemoCheckRc(g_hPhysDevs.hBurst, result);

}



void DemoEnablePhysicalDevices(void)
/*******************************************************************************
  @func
    Enable Configured Physical Devices
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  DemoLogf("In DemoEnablePhysicalDevices\n");
  result = OtxDrvEnable(g_hPhysDevs.hBoardController0,     OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  for (i = 0; i < DEMO_DSP_MAX; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i],            OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }
    

  // by enabling the Burst device, we start the process of transferring data
  // from the hardware to the driver via DMA.  The data is stored in the Burst
  // device's buffer for later processing to connected logical devices.
  DemoLogf("Enable g_hPhysDevs.hBurst\n");
  result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
  DemoCheckRc(g_hPhysDevs.hBurst, result);


}


/*******************************************************************************
  @func
    Create the needed Logical Devices

  @comm
    The logical devices are different from the phyiscal devices in that the
    the number of logical devices is not fixed. More of them can be requested
    by the user and allocated by the system if needed.
*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  nDsp;
  OTX_INT32  nLi;

  for (nLi = 0; nLi < DEMO_LI_MAX; nLi++) {
    
    nDsp = nLi;
    
    DemoLogf("Creating Logical Devices on Dsp%d\n", nDsp);
    
    if (g_demoEnv.outFileName) {
      // For debug prints
      result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nLi], OTX_LDEVICE_USER_APPLICATION, g_hEventQueue, nLi, &g_hLogicDevs.hUserApp[nLi]); DemoCheckRc(g_hPhysDevs.hDsp[nLi], result);
      DemoLogf("Created User App device [%d], Handle = 0x%X (for debug printfs)\n", nLi, g_hLogicDevs.hUserApp[nLi]);
    }
    
    // Create Pots Li Logical Devices
    DemoLogf("Create Logical Device OTX_LDEVICE_CST_LI_POTS #%d (nDsp=%d)\n", nLi, nDsp);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_CST_LI_POTS, g_hEventQueue, nLi, &g_hLogicDevs.hPotsLi[nLi]); DemoCheckRc(g_hLogicDevs.hPotsLi[nLi], result);

    // Create Echo Cancel Device Logical Devices
    DemoLogf("Creating Echo Cancel Device device [%d]\n", nDsp);
    result = OtxDrvCreateLogicalDevice( g_hPhysDevs.hDsp[nDsp], OTX_LDEVICE_CST_ECHO_CANCEL, g_hEventQueue, nDsp, &g_hLogicDevs.hEchoCancel[nDsp]); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
    DemoLogf("Created Echo Cancel Device device [%d], Handle = 0x%X \n", nDsp, g_hLogicDevs.hEchoCancel[nDsp]);
  } 
  
}


void DemoConnectDevices(
/*******************************************************************************
  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.
*******************************************************************************/
  void
)
{
  OTX_INT32 nDsp;
  OTX_RESULT result;
  OTX_UINT32 iHighwayNo;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_INT32 nTs;


  // Create the Pipes (one for each direction) between the Burst Device and the DSPs
  for (nDsp = 0; nDsp < DEMO_LI_MAX; nDsp++) {

    // Burst highway number [0]
    iHighwayNo = 0;

    // PIPE CONNECTIONS BETWEEN PHYSICAL DEVICES
    // -----------------------------------------

    //pipe = hDaaRx, which contains
    //received samples from the DAA device (signal with hybrid echo).

    //  INPUT pipe end:  PotsLi device     (ConnPoint = -1) DSPHwyTs=0 
    //  OUTPUT pipe end: EchoCancel device (ConnPoint = 0)  DSPHwyTs=1
    result = OtxDrvCreatePipe(64, &g_hPipes.hDaaRx[nDsp]); 
    DemoCheckRc(g_hPipes.hDaaRx[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 0;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hDaaRx[nDsp], OTX_PIPE_INPUT, g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32); 
    DemoCheckRc(g_hPipes.hDaaRx[nDsp], result);

#ifndef DEMO_EC_DEBUG
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 1;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hDaaRx[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32); 
    DemoCheckRc(g_hPipes.hDaaRx[nDsp], result);
#endif // DEMO_EC_DEBUG

    //pipe = hDaaTx, which contains
    //the reference/prompt samples to be output to the DAA device. The samples are
    //provided by the burst device.
    
    //  INPUT pipe end:  Burst device                       BurstHwyTs=0
    //  OUTPUT pipe end: PotsLi device     (ConnPoint = -1) DSPHwyTs=0

    result = OtxDrvCreatePipe(64, &g_hPipes.hDaaTx[nDsp]);
    DemoCheckRc(g_hPipes.hDaaTx[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 0;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hDaaTx[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hBurst, iHighwayNo, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hDaaTx[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
#ifdef DEMO_EC_DEBUG
    nTs = 1;
#else
    nTs = 0;
#endif
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hDaaTx[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32); 
    DemoCheckRc(g_hPipes.hDaaTx[nDsp], result);



    //pipe = hBurstTx, which contains
    //the delayed reference/prompt samples to be the Near Input to the Echo
    //Cancel device. The samples are provided by the burst device. The application
    //will buffer (delay) these samples by 150ms. Other than the delay the data
    //that travels through this pipe is the same as for the hDaaTx pipe.
    
    //  INPUT pipe end:  Burst device                       BurstHwyTs=1
    //  OUTPUT pipe end: EchoCancel device (ConnPoint = 1)  DSPHwyTs=2
    result = OtxDrvCreatePipe(64, &g_hPipes.hBurstTx[nDsp]);
    DemoCheckRc(g_hPipes.hBurstTx[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 1;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hBurstTx[nDsp], OTX_PIPE_INPUT,  g_hPhysDevs.hBurst, iHighwayNo, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hBurstTx[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 2;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hBurstTx[nDsp], OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32); 
    DemoCheckRc(g_hPipes.hBurstTx[nDsp], result);


    //pipe = hBurstRx, which contains
    //the samples from the Echo Cancel device (after the hybrid echo has been
    //remove/cancelled).
    
    //  INPUT pipe end:  EchoCancel device (ConnPoint = -1) DSPHwyTs=10
    //  OUTPUT pipe end: Burst device                       BurstHwyTs=0
    result = OtxDrvCreatePipe(64, &g_hPipes.hBurstRx[nDsp]);
    DemoCheckRc(g_hPipes.hBurstRx[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 10;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hBurstRx[nDsp], OTX_PIPE_INPUT, g_hPhysDevs.hDsp[nDsp],  -1, sqTimeSlotMask, 32); 
    DemoCheckRc(g_hPipes.hBurstRx[nDsp], result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 0;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(g_hPipes.hBurstRx[nDsp], OTX_PIPE_OUTPUT,  g_hPhysDevs.hBurst, iHighwayNo, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hBurstRx[nDsp], result);

    // CONNECT LOGICAL DEVICES WITHIN THE DSP
    // --------------------------------------

    // Device[hPotsLi] (CP = -1) --> Pipe[hDaaRx] (CP = -1) 
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hPotsLi[nDsp], -1, g_hPipes.hDaaRx[nDsp], -1); 
    DemoCheckRc(g_hLogicDevs.hPotsLi[nDsp], result);
    // Pipe[hDaaTx] (CP = -1) --> Device[hPotsLi] (CP = -1) 
#ifndef DEMO_EC_DEBUG
    result = OtxDrvConnectLogicalDevice(g_hPipes.hDaaTx[nDsp], -1, g_hLogicDevs.hPotsLi[nDsp], -1); 
    DemoCheckRc(g_hPipes.hDaaTx[nDsp], result);
#endif

    // Pipe[hDaaRx] (CP = -1) -> Device[hEchoCancel] (CP = 0 (Near End input))
    // Pipe[] (CP = -1) -> Device[hEchoCancel] (CP = 1 (Far End input))
    // Device[hEchoCancel] (CP = -1) -> Pipe[] (CP = -1) 
#ifdef DEMO_EC_DEBUG
    result = OtxDrvConnectLogicalDevice(g_hPipes.hDaaTx[nDsp], -1, g_hLogicDevs.hEchoCancel[nDsp], OTX_CST_ECHO_CANCEL_NEAR_END_PIPE); 
#else
    result = OtxDrvConnectLogicalDevice(g_hPipes.hDaaRx[nDsp], -1, g_hLogicDevs.hEchoCancel[nDsp], OTX_CST_ECHO_CANCEL_NEAR_END_PIPE); 
#endif
    DemoCheckRc(g_hPipes.hDaaRx[nDsp], result);
    result = OtxDrvConnectLogicalDevice(g_hPipes.hBurstTx[nDsp], -1, g_hLogicDevs.hEchoCancel[nDsp], OTX_CST_ECHO_CANCEL_FAR_END_PIPE); 
    DemoCheckRc(g_hPipes.hBurstTx[nDsp], result);
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hEchoCancel[nDsp], -1, g_hPipes.hBurstRx[nDsp], -1); 
    DemoCheckRc(g_hLogicDevs.hEchoCancel[nDsp], result);
  }

#if 1 // For additional debugging
  {
    OTX_HANDLE hPipeNe;
    OTX_HANDLE hPipeFe;
    OTX_INT32 nHwy;
    
    // Also connect the DAA samples (the EC Near End Input) to the burst via a TSS connection
    result = OtxDrvCreatePipe(64, &hPipeNe);
    DemoCheckRc(hPipeNe, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
#ifdef DEMO_EC_DEBUG
    nTs = 0;
    nHwy = 8;
#else
    nTs = 0;
    nHwy = 2;
#endif
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeNe, OTX_PIPE_INPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeNe, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 1;
    nHwy = 8;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeNe, OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeNe, result);

    // Also connect the EC Far End Input to the burst via a TSS connection
    result = OtxDrvCreatePipe(64, &hPipeFe);
    DemoCheckRc(hPipeFe, result);

    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 1;
    nHwy = 8;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeFe, OTX_PIPE_INPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeFe, result);
    
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    nTs = 2;
    nHwy = 8;
    sqTimeSlotMask[nTs] = 0xFF;
    result = OtxDrvConnectPipe(hPipeFe, OTX_PIPE_OUTPUT, g_hPhysDevs.hTss,  nHwy, sqTimeSlotMask, 32); 
    DemoCheckRc(hPipeFe, result);

  }
#endif

}


void DemoInitializeLogicalDevices(
/*******************************************************************************
  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.
*******************************************************************************/
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  DemoLogf("In DemoInitializeLogicalDevices\n");


  DemoLogf("Configuring Lis for country set=");
  switch (g_demoEnv.nSetId) {
  case OTX_CST_POTS_CFG_SET_USA:
    DemoLogf("USA");
    break;
  case OTX_CST_POTS_CFG_SET_GERMANY: 
    DemoLogf("Germany");
    break;
  case OTX_CST_POTS_CFG_SET_UK:
    DemoLogf("United Kingdom");
    break;
  case OTX_CST_POTS_CFG_SET_FRANCE:
    DemoLogf("France");
    break;
  case OTX_CST_POTS_CFG_SET_AUSTRALIA:
    DemoLogf("Australia");
    break;
  case OTX_CST_POTS_CFG_SET_JAPAN:
    DemoLogf("Japan");
    break;
  default:
    DemoLogf("UNKNOWN");
    break;
  }
  DemoLogf("\n");  

  for (i = 0; i < DEMO_LI_MAX; i++) {

    // Initialize the POTS LI logical devices
    result = OtxCstPotsInit(
      g_hLogicDevs.hPotsLi[i],
      g_demoEnv.law,
      g_demoEnv.nSetId
      );
    DemoCheckRc(g_hLogicDevs.hPotsLi[i], result);

    // Initialize the Echo Cancel Device
    result = OtxCstEchoCancelInit(
      g_hLogicDevs.hEchoCancel[i], 
      g_demoEnv.law, 
      OTX_CST_ECHO_CANCEL_32_MS); 
    DemoCheckRc(g_hLogicDevs.hEchoCancel[i], result);
  }
}




void DemoEnableLogicalDevices(void)
/*******************************************************************************
  @func
    Enable Configured Logical Devices
*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32  i;

  DemoLogf("In DemoEnableLogicalDevices\n");

  for (i = 0; i < DEMO_LI_MAX; i++) {

    result = OtxDrvEnable(g_hLogicDevs.hPotsLi[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hPotsLi[i], result);

    /* Done later at the 'R' command
    result = OtxDrvEnable(g_hLogicDevs.hEchoCancel[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hEchoCancel[i], result);
    */
  }

}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX [%d.%d])\n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result));
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    DemoExit(1);
  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
    
}


/*******************************************************************************
  @func
    Function that keeps reseting the Watchdog timer on the board
*******************************************************************************/
void  *DemoWatchDogFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_BOOL bEnabled;

  g_bWatchDogThreadActive = OTX_TRUE;

  DemoLogf("DemoWatchDogFunction Thread started\n");

  while (g_bWatchDogThreadActive) {

    result = OtxDrvGetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_ENABLED,  &bEnabled); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    if (!bEnabled) {
      DemoLogf("*********** Watchdog Timer timed out. Application must be restarted *********\n");
      OsaThreadSleep(3000); // in milliseconds
      continue;
    }

    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_RESET,  OTX_TRUE /* Any value will work */); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    OsaThreadSleep(100); // in milliseconds
  }

  DemoLogf("DemoWatchDogFunction Thread terminating\n");
  return 0;

}




/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;

}


/*******************************************************************************
  @func
    Only used by the USER_APPLICATION device for debugging of DSP program code.
*******************************************************************************/
void DemoPrintCSTEvent(OtxEventDataS *pEventData) 
{
  OTX_RESULT otxRc;
  char msgStr[256];
  
  if (pEventData->m_nCode == OTX_USER_APPLICATION_EC_READ_DATA) {
    OTX_UINT8  buf[8192];
    OTX_UINT32 nBytesReturned;
    OTX_UINT32 nIoControlCode;      
    
    otxRc = OtxDspReadData(pEventData->m_hDevice, OTX_TASK_SYNC, buf, 8192, &nBytesReturned, &nIoControlCode);
    if (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
      switch(nIoControlCode) {
      case 0x2ff:
        {
          char strBuf[1024];
          int k;
          sprintf(strBuf, "Len=%d :",nBytesReturned);
          for(k=0;k<(int)nBytesReturned;k+=2){
            sprintf(&strBuf[0] + strlen(strBuf), "%c",buf[k]);
          }
          sprintf(&strBuf[0] + strlen(strBuf), "\n");
          DemoLogf(strBuf);
        }
        break;
        
      case 0x0200 /*DEMO1_NOTIFY_DEBUG_DATA*/:
        {
          DemoLogf("DEMO1_NOTIFY_DEBUG_DATA (nRequestId=%d, nBytes=%d)\n", pEventData->m_nRequestId, nBytesReturned);
        }
        break;

      case 0x02FE:
        {
          static int counter = 0;

          if (counter++ % 5000 == 0) {
            DemoLogf("Got data\n");
          }

          if (g_fStreamFile) {
            fwrite((OTX_UINT8 *)buf, nBytesReturned, sizeof(OTX_UINT8), g_fStreamFile);
          }

        }
        break;
      default:
        DemoLogf("Unknown IoControlCode received in OtsDspReadData (nIoControlCode = %d, nBytesReturned=%d, nRequestId=%d)\n", nIoControlCode, nBytesReturned, pEventData->m_nRequestId);
        break;
      }
    } else {
      OTX_RESULT errorCode = otxRc;
      otxRc = OtxDrvResultCode2String(pEventData->m_hDevice, errorCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("User App[%d] Error: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("User App[%d] Invalid Error Code: %ld\n", pEventData->m_nSourceId, errorCode);
      }
      
    }
    
  } else {
    
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("User App[%d] Event: %s (nRequestId=%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nRequestId);
    } else {
      DemoLogf("User App[%d] Invalid Event: %ld (nRequestId=%d)\n", pEventData->m_nSourceId, pEventData->m_nCode, pEventData->m_nRequestId);
    }
  }
}




/*******************************************************************************
  @func
    Burst Sender Function
*******************************************************************************/
void DemoBurstSender(OTX_UINT32 nSourceId)
{
  OTX_RESULT  otxRc;
  OTX_UINT32  ui32;
  OTX_INT32 nTsToSendDirect = 0;
  OTX_INT32 nTsToSendDelayed = 1;
  int cbSingleTsBuf;
  int byteCount = 0;
  int x;

  // validate parameters
  if (nSourceId != 0) {
    DemoLogf("DemoBurstSender: Invalid nSourceId(%d)!\n", nSourceId);
    return;
  }


  if (!g_sBurstDevs.mBurstSender.pTxFile) {
    
    if ((g_sBurstDevs.mBurstSender.pTxFile = fopen(DEMO_EC_FE_INPUT_FILE, "rb")) == NULL) {
      DemoLogf("DemoBurstSender: Could not open file %s. Disabling the burst device\n", DEMO_EC_FE_INPUT_FILE);
      OtxDrvDisable(g_hPhysDevs.hBurst, 5000, 2000);
      return;
    }

#ifdef DEMO_EC_DEBUG
    if ((g_sBurstDevs.mBurstSender.pTxFile2 = fopen(DEMO_EC_ECHO_INPUT_FILE, "rb")) == NULL) {
      DemoLogf("DemoBurstSender: Could not open file %s. Disabling the burst device\n", DEMO_EC_ECHO_INPUT_FILE);
      OtxDrvDisable(g_hPhysDevs.hBurst, 5000, 2000);
      return;
    }
#endif
    
#ifndef DEMO_EC_DEBUG
    g_sBurstDevs.mBurstSender.bFileLoop = OTX_TRUE;
#endif

    g_sBurstDevs.mBurstSender.bActive = OTX_TRUE;

    DemoLogf("Burst Sender started\n");

    memset(&g_sBurstDevs.mBurstSender.sqBufCurr[0], DEMO_U_LAW_SILENCE, sizeof(g_sBurstDevs.mBurstSender.sqBufCurr) / sizeof(g_sBurstDevs.mBurstSender.sqBufCurr[0]));
    memset(&g_sBurstDevs.mBurstSender.sqBufPrev[0], DEMO_U_LAW_SILENCE, sizeof(g_sBurstDevs.mBurstSender.sqBufPrev) / sizeof(g_sBurstDevs.mBurstSender.sqBufPrev[0]));

  }
  

  if (!g_sBurstDevs.mBurstSender.bActive) {
    // we're not active, so just return
    return;
  }

  if (!g_sBurstDevs.mBurstSender.pTxFile) {
    // No file to send, so just return
    return;
  }


  // Prime the transmit buffer with DEMO_U_LAW_SILENCE (0xFF)
  memset(&g_sBurstDevs.mBurstSender.sqTxBuf[0], DEMO_U_LAW_SILENCE, sizeof(g_sBurstDevs.mBurstSender.sqTxBuf) / sizeof(g_sBurstDevs.mBurstSender.sqTxBuf[0]));

  // Copy the previous buffer so we can use its samples for the delayed sample buffer
  memcpy(&g_sBurstDevs.mBurstSender.sqBufPrev[0], &g_sBurstDevs.mBurstSender.sqBufCurr[0], sizeof(g_sBurstDevs.mBurstSender.sqBufPrev) / sizeof(g_sBurstDevs.mBurstSender.sqBufPrev[0]));
  

#ifdef DEMO_EC_DEBUG
  cbSingleTsBuf = fread(
    &g_sBurstDevs.mBurstSender.sqBufCurr,
    sizeof(g_sBurstDevs.mBurstSender.sqBufCurr[0]),
    sizeof(g_sBurstDevs.mBurstSender.sqBufCurr) / sizeof(g_sBurstDevs.mBurstSender.sqBufCurr[0]),
    g_sBurstDevs.mBurstSender.pTxFile2
    );
  // Dispurse the single TS to be transmitted in the whole transmit buffer 
  // for the SendDirect timeslot on the burst highway
  byteCount = 0;
  for ( x = nTsToSendDirect; (byteCount < cbSingleTsBuf) && (x < DEMO_BURST_BUF_SIZE); x += 32) {
    g_sBurstDevs.mBurstSender.sqTxBuf[x] = g_sBurstDevs.mBurstSender.sqBufCurr[byteCount++];
  }
#endif // DEMO_EC_DEBUG
  
  cbSingleTsBuf = fread(
    &g_sBurstDevs.mBurstSender.sqBufCurr,
    sizeof(g_sBurstDevs.mBurstSender.sqBufCurr[0]),
    sizeof(g_sBurstDevs.mBurstSender.sqBufCurr) / sizeof(g_sBurstDevs.mBurstSender.sqBufCurr[0]),
    g_sBurstDevs.mBurstSender.pTxFile
    );
  
  
  // Dispurse the single TS to be transmitted in the whole transmit buffer 
  // for the SendDirect timeslot on the burst highway

#ifndef DEMO_EC_DEBUG
  byteCount = 0;
  for ( x = nTsToSendDirect; (byteCount < cbSingleTsBuf) && (x < DEMO_BURST_BUF_SIZE); x += 32) {
    g_sBurstDevs.mBurstSender.sqTxBuf[x] = g_sBurstDevs.mBurstSender.sqBufCurr[byteCount++];
  }
#endif
  
  assert(byteCount == cbSingleTsBuf);
  g_sBurstDevs.mBurstSender.cbTxBuf = cbSingleTsBuf * 32;

#define DEMO_DELAY_SAMPLES (32)  // 32 is 8 ms
  // Delay the samples 8 ms (32 samples) in the delayed buffer for the SendDelayed timeslot on the burst highway.
  // First copy 32 samples from the previous buffer, then the rest from the current buffer
  byteCount = 0;
  for ( x = nTsToSendDelayed; (byteCount < cbSingleTsBuf) && (x < DEMO_BURST_BUF_SIZE); x += 32) {
#if 1
    if (byteCount < DEMO_DELAY_SAMPLES) {
      g_sBurstDevs.mBurstSender.sqTxBuf[x] = g_sBurstDevs.mBurstSender.sqBufPrev[(sizeof(g_sBurstDevs.mBurstSender.sqBufCurr) / sizeof(g_sBurstDevs.mBurstSender.sqBufCurr[0]) - DEMO_DELAY_SAMPLES) + byteCount++];
    } else {
      g_sBurstDevs.mBurstSender.sqTxBuf[x] = g_sBurstDevs.mBurstSender.sqBufCurr[byteCount++ - DEMO_DELAY_SAMPLES];
    }
#else // Don't delay
    g_sBurstDevs.mBurstSender.sqTxBuf[x] = g_sBurstDevs.mBurstSender.sqBufCurr[byteCount++];
#endif
  }
  
  if (!feof(g_sBurstDevs.mBurstSender.pTxFile)) {
    assert(g_sBurstDevs.mBurstSender.cbTxBuf == DEMO_BURST_BUF_SIZE);
  } else {
    DemoLogf("DemoBurstSender[%d]: Reached EOF\n", nSourceId);
  }
  
  
  if (g_sBurstDevs.mBurstSender.cbTxBuf > 0) {
    
    // echo if we're in verbose mode
    if (g_sBurstDevs.mBurstSender.bVerbose) {
      int x;
      
      DemoLogf("DemoBurstSender[%d]: Transmitting %d bytes\n", nSourceId, g_sBurstDevs.mBurstSender.cbTxBuf);
      
      for (x = 0; (x < 32) && (x < (OTX_INT32) sizeof(g_sBurstDevs.mBurstSender.sqTxBuf)); x++) {
        // display the highway's currentLi's data
        if ((x % 8) == 0) DemoLogf("\n");
        DemoLogf("%.2X ", g_sBurstDevs.mBurstSender.sqTxBuf[x]);
      }
      DemoLogf("\n");
      DemoLogf("\n");
    }

    
    ui32 = g_sBurstDevs.mBurstSender.cbTxBuf;
    otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, nSourceId, g_sBurstDevs.mBurstSender.sqTxBuf, &ui32);
    DemoCheckRc(g_hPhysDevs.hBurst, otxRc);
    
    // check the amount buffered w/ how much we requested to be buffered
    //  (since we set the trigger amount to the buffer size, there should
    //    have been room to buffer this, so these two should be the same).

    if (!(ui32 == g_sBurstDevs.mBurstSender.cbTxBuf)) {
      DemoLogf("ui32(%d) != g_sBurstDevs.mBurstSender.cbTxBuf(%d)\n", ui32, g_sBurstDevs.mBurstSender.cbTxBuf);
    }
    
  } else {
    
    if (g_sBurstDevs.mBurstSender.bFileLoop) {
      if (g_sBurstDevs.mBurstSender.bVerbose) {
        DemoLogf("DemoBurstSender[%d]: Transmission complete. Starting file from the beginning.\n", nSourceId);
      }
      rewind(g_sBurstDevs.mBurstSender.pTxFile);
    } else {
      DemoLogf("DemoBurstSender[%d]: Transmission complete. Closing file\n", nSourceId);
      fclose(g_sBurstDevs.mBurstSender.pTxFile);
      g_sBurstDevs.mBurstSender.pTxFile = NULL;
#ifdef DEMO_EC_DEBUG
      fclose(g_sBurstDevs.mBurstSender.pTxFile2);
      g_sBurstDevs.mBurstSender.pTxFile2 = NULL;
#endif
      g_sBurstDevs.mBurstSender.bActive = OTX_FALSE;
    }
  }
  
  return;
}

/*******************************************************************************
  @func
    Burst Receiver Function
*******************************************************************************/
void DemoBurstReceiver(OTX_UINT32 nSourceId)
{
  OTX_RESULT    otxRc;
  OTX_UINT8     buf[DEMO_BURST_BUF_SIZE];
  OTX_UINT32    ui32;
  OTX_UINT64    nFrameRefId;
  OTX_DATETIME  dtFrameRefTime;

  // validate parameters
  if (nSourceId < 0 || nSourceId > 2) {
    DemoLogf("DemoBurstReceiver: Invalid nSourceId(%d)...\n", nSourceId);
    return;
  }

  if (!g_sBurstDevs.mBurstReceiver.bActive) {
    // we're not active, so just return
    return;
  }

  // we want the frame reference id and time
  ui32 = sizeof(buf);
  otxRc = OtxBurstReadDataEx(g_hPhysDevs.hBurst, nSourceId, buf, &ui32, &nFrameRefId, &dtFrameRefTime);
  DemoCheckRc(g_hPhysDevs.hBurst, otxRc);

  // check the amount buffered w/ how much we requested to be buffered
  //  (since we set the trigger amount to the buffer size, there should
  //    have been room to buffer this, so these two should be the same).
  assert(ui32 == sizeof(buf));

  // echo if we're in verbose mode
  if (g_sBurstDevs.mBurstReceiver.bVerbose) {
    int x;

    DemoPrintTimeStamp(dtFrameRefTime);

    // note: we only display the least significant 32 bits of the frame reference id.
    DemoLogf("BurstReceiver[%d] (len=%.5d) ID:%08x ", nSourceId, sizeof(buf), (OTX_UINT32) (nFrameRefId & 0xffffffff));

    for (x = 0; (x < 64) && (x < (OTX_INT32) sizeof(buf)); x++) {
      // display the highway's currentLi's data
      if ((x % 8) == 0) DemoLogf("\n");
      DemoLogf("%.2X ", buf[x]);
    }
    DemoLogf("\n");
    DemoLogf("\n");
  }

  // write received data to output files
  if (g_sBurstDevs.mBurstReceiver.pRxFile) {
    int cbWritten;
    
    // TS0 is connected to EC Output
    OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE/32];
    int byteCount = 0;
    int x;
    
    byteCount = 0;
    for (x = g_sBurstDevs.mBurstReceiver.nTsToRecord; x < (OTX_INT32) sizeof(buf); x += 32) {
      singleTsBuf[byteCount++] = buf[x];
    }
    cbWritten = fwrite(singleTsBuf, sizeof(singleTsBuf[0]), byteCount, g_sBurstDevs.mBurstReceiver.pRxFile);
    if (cbWritten != byteCount) {
      DemoLogf("BurstReceiver[%d] Error: Could not write data to file 1.  (byteCount=%d, cbWritten=%d) \n", nSourceId, byteCount, cbWritten);
    }
    
  }
  if (g_sBurstDevs.mBurstReceiver.pRxFile2) {
    int cbWritten;
    
    // TS1 is connected to NE input
    OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE/32];
    int byteCount = 0;
    int x;
    
    byteCount = 0;
    for (x = g_sBurstDevs.mBurstReceiver.nTsToRecord+1; x < (OTX_INT32) sizeof(buf); x += 32) {
      singleTsBuf[byteCount++] = buf[x];
    }
    cbWritten = fwrite(singleTsBuf, sizeof(singleTsBuf[0]), byteCount, g_sBurstDevs.mBurstReceiver.pRxFile2);
    if (cbWritten != byteCount) {
      DemoLogf("BurstReceiver[%d] Error: Could not write data to file 2.  (byteCount=%d, cbWritten=%d) \n", nSourceId, byteCount, cbWritten);
    }
    
  }
  if (g_sBurstDevs.mBurstReceiver.pRxFile3) {
    int cbWritten;
    
    // TS2 is connected FE input
    OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE/32];
    int byteCount = 0;
    int x;
    
    byteCount = 0;
    for (x = g_sBurstDevs.mBurstReceiver.nTsToRecord+2; x < (OTX_INT32) sizeof(buf); x += 32) {
      singleTsBuf[byteCount++] = buf[x];
    }
    cbWritten = fwrite(singleTsBuf, sizeof(singleTsBuf[0]), byteCount, g_sBurstDevs.mBurstReceiver.pRxFile3);
    if (cbWritten != byteCount) {
      DemoLogf("BurstReceiver[%d] Error: Could not write data to file 3.  (byteCount=%d, cbWritten=%d) \n", nSourceId, byteCount, cbWritten);
    }
    
  }
}

void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];
  
  //DemoPrintTimeStamp(pEventData->m_nTimeStamp);
  
  switch (pEventData->m_eDeviceType) {
    
  case OTX_LDEVICE_USER_APPLICATION:
    // Events from the DspDemo1.out logical devices
    DemoPrintCSTEvent(pEventData);
    break;
    
  case OTX_DEVICE_DSP:
    {
      // Non-logical device related event from the DSPs
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("DSP[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;
    
  case OTX_LDEVICE_CST_DTMF_DIALER:
    {
      switch (pEventData->m_nCode) {
      case OTX_CST_DTMF_DIALER_EC_COMPLETE:
        DemoLogf("DtmfDialer[%d] Completed DTMF dialing of %d digits\n", pEventData->m_nSourceId, pEventData->m_nParam);
        break;
      default:
        DemoLogf("DtmfDialer[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
        break;
      }
    }
    break;
    
    
  case OTX_LDEVICE_CST_DTMF_DETECTOR:
    {
      // Event from Dtmf Detector
      switch (pEventData->m_nCode) {
      case OTX_CST_DTMF_DETECTOR_EC_MTONESTART:
        {
          DemoLogf("DtmfDetector[%d] MTONESTART, DTMF digit %c\n", pEventData->m_nSourceId, DemoDtmfParamToChar(pEventData->m_nParam));
        }
        break;
        
      case OTX_CST_DTMF_DETECTOR_EC_MTONE:
        {
          DemoLogf("DtmfDetector[%d] MTONE, DTMF tone confirmed\n", pEventData->m_nSourceId);
        }
        break;
        
      case OTX_CST_DTMF_DETECTOR_EC_MSIGNALEND:
        {
          DemoLogf("DtmfDetector[%d] MSIGNALEND, DTMF digit %c\n", pEventData->m_nSourceId, DemoDtmfParamToChar(pEventData->m_nParam));
        }
        break;
        
      case OTX_CST_DTMF_DETECTOR_EC_MFAIL:
        {
          DemoLogf("DtmfDetector[%d] MFAIL\n", pEventData->m_nSourceId);
        }
        break;
        
      case OTX_CST_DTMF_DETECTOR_EC_MPATTERNEND:
        {
          DemoLogf("DtmfDetector[%d] MPATTERNEND\n", pEventData->m_nSourceId);
        }
        break;
        
      default:
        DemoLogf("DtmfDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
        break;
        
      }
    }
    break;
    
  case OTX_LDEVICE_CST_LI_POTS:
    {
      // Event from Pots Li
      switch (pEventData->m_nCode) {

      case OTX_CST_POTS_RING_SIGNAL_DETECTED:
        DemoLogf("PotsLi[%d] A ring signal has been detected\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_END_RING_SIGNAL_DETECTED:
        DemoLogf("PotsLi[%d] End of ring signal has been detected\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_AUTO_END_RING_SIGNAL_DETECTED:
        DemoLogf("PotsLi[%d] End of ring signal due to offhook has been detected\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_TIP_RING_REVERSAL_DETECTED:
        DemoLogf("PotsLi[%d] Tip/Ring reversal detected. The other end hung up upon us.\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_CALLER_ID_DETECTED:
        {
          char szCallerId[40];

          if (pEventData->m_nParam == OTX_CST_CALLER_ID_EC_NOERROR) {
            otxRc = OtxDrvGetAttributeValueSz(pEventData->m_hDevice, OTX_ATTR_CST_POTS_CALLER_ID, szCallerId, 40); 
            if (OTX_SUCCESS(otxRc)) {
              DemoLogf("PotsLi[%d] Caller Id = \"%s\"\n", pEventData->m_nSourceId, szCallerId);
            } else if (OTX_RESULT_CODE(otxRc) == OTX_RES_BUSY) {
              DemoLogf("PotsLi[%d] Caller Id Overrun. Caller Id string lost. Reseting the condition\n", pEventData->m_nSourceId);
              OtxDrvSetAttributeValueSz(pEventData->m_hDevice, OTX_ATTR_CST_POTS_CALLER_ID, NULL); 
            } else {
              DemoLogf("PotsLi[%d] Unexpected return code from OtxDrvGetAttributeValueSz (otxRc=0x%X)\n", pEventData->m_nSourceId, otxRc);
            }
          } else {
            DemoLogf("PotsLi[%d] Caller Id detection failure. nCode=%d\n", pEventData->m_nSourceId, pEventData->m_nParam);
          }

        }
        break;
      case OTX_CST_POTS_OFF_HOOK_COMPLETE:
        assert(pEventData->m_nSourceId < DEMO_LI_MAX);
        g_sPotsDevs[pEventData->m_nSourceId].mPotsLi.bOffHook = OTX_TRUE;
        DemoLogf("PotsLi[%d] Off-Hook command complete\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_ON_HOOK_COMPLETE:
        assert(pEventData->m_nSourceId < DEMO_LI_MAX);
        g_sPotsDevs[pEventData->m_nSourceId].mPotsLi.bOffHook = OTX_FALSE;
        DemoLogf("PotsLi[%d] On-Hook command complete\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_ON_HOOK_LINE_MON_COMPLETE:
        assert(pEventData->m_nSourceId < DEMO_LI_MAX);
        g_sPotsDevs[pEventData->m_nSourceId].mPotsLi.bOffHook = OTX_FALSE;
        DemoLogf("PotsLi[%d] On-Hook Line Monitor mode command complete\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_SENSE_LINE_COMPLETE:
        assert(pEventData->m_nSourceId < DEMO_LI_MAX);
        if (g_sPotsDevs[pEventData->m_nSourceId].mPotsLi.bOffHook) {
          // We are in the OffHook state so line sensing will return loop current (3mA per bit)
          DemoLogf("PotsLi[%d] Sense line command complete. Loop current = %d mA (%d) above minimum\n", pEventData->m_nSourceId, 3 * pEventData->m_nParam, pEventData->m_nParam);
        } else {
          double nVal = 2.75 * pEventData->m_nParam;
          // We are in the OnHook state so line sensing will return line voltage (2.75V per bit)
          DemoLogf("PotsLi[%d] Sense line command complete. Line voltage = %2.1f V  (%d)\n", pEventData->m_nSourceId, (double) nVal, pEventData->m_nParam);
        }

        DemoLogf("PotsLi[%d] Sense line command complete Val=%d\n", pEventData->m_nSourceId, pEventData->m_nParam);
        break;
      case OTX_CST_POTS_PULSE_GEN_COMPLETE:
        DemoLogf("PotsLi[%d] Pulse Generator command complete\n", pEventData->m_nSourceId);
        break;
      case OTX_CST_POTS_READ_REG_COMPLETE:
        DemoLogf("PotsLi[%d] ReadReg command complete Val=%d\n", pEventData->m_nSourceId, pEventData->m_nParam);
        break;
      case OTX_CST_POTS_WRITE_REG_COMPLETE:
        DemoLogf("PotsLi[%d] WriteReg command complete\n", pEventData->m_nSourceId);
        break;
      default:
        // Unexpected event from this device
        DemoLogf("PotsLi[%d] Unexpected Event (%d)\n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;
    
  case OTX_DEVICE_BURST:
    {
      switch (pEventData->m_nCode) {
      case OTX_BURST_RECEIVER_EC_DATA_AVAILABLE:
        DemoBurstReceiver(pEventData->m_nRequestId);
        break;
      case OTX_BURST_SENDER_EC_BUF_DEPLETED:
        DemoBurstSender(pEventData->m_nRequestId);
        break;
      default:
        otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
        if (OTX_SUCCESS(otxRc)) {
          DemoLogf("Burst[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
        } else {
          DemoLogf("Burst[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
        }
        break;
      }
      
    }
    break;
    
  default:
    {
      // Unhandled event
      DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    }
    break;
    
  }
  
}

/*******************************************************************************

  @func

  Convert a DTMF parameter to a digit (char).

  @comm

*******************************************************************************/
static char DemoDtmfParamToChar(int nParam) {
  switch (nParam) {
  case 0:  // 0  (941Hz/1336Hz)
  case 1:  // 1  (697Hz/1209Hz)
  case 2:  // 2  (697Hz/1336Hz)
  case 3:  // 3  (697Hz/1477Hz)
  case 4:  // 4  (770Hz/1209Hz)
  case 5:  // 5  (770Hz/1336Hz)
  case 6:  // 6  (770Hz/1477Hz)
  case 7:  // 7  (852Hz/1209Hz)
  case 8:  // 8  (852Hz/1336Hz)
  case 9:  // 9  (852Hz/1477Hz)
    return '0' + nParam;
  case 10: // A  (697Hz/1633Hz)
  case 11: // B  (770Hz/1633Hz)
  case 12: // C  (852Hz/1633Hz)
  case 13: // D  (941Hz/1633Hz)
    return 'A' - 10 + nParam;
  case 14: // *  (941Hz/1209Hz)
    return '*';
    break;
  case 15: // #  (941Hz/1477Hz)
    return '#';
    break;
  default: // Error
    DemoLogf("Unrecognized DTMF tone parameter (%d)\n", nParam);
    return '-';
  }
}





/*******************************************************************************

  @func

*******************************************************************************/
OTX_BOOL DemoIsBurstDeviceEnabled()
{
  OTX_RESULT result;
  OtxDeviceStateE  eDevState;

  result = OtxDrvGetState(g_hPhysDevs.hBurst, &eDevState); 
  DemoCheckRc(g_hPhysDevs.hBurst, result);

  if (eDevState == OTX_DEVICE_WORKING) {
    return OTX_TRUE; // The burst device is enabled
  } else {
    return OTX_FALSE; // The burst device is not enabled
  }
}


OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************
  @func
    Processes User Commands
*******************************************************************************/
{
  OTX_RESULT        result;
  static OTX_INT32  nCurrentLi = 0;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd   Synopsis\n");
    DemoLogf("===   ===========================================================\n");
    DemoLogf(" ?    Display Help text\n");
    DemoLogf(" h    Ditto\n");
    DemoLogf(" q    Quit\n");

    DemoLogf(" 0-5  Select active Line Interface\n");
    DemoLogf(" f    Set active Pots Li Off-Hook\n");
    DemoLogf(" n    Set active Pots Li On-Hook\n");
    DemoLogf(" l    Flash/hook on the active Pots Li\n");
    DemoLogf(" r    Read a Register from the DAA\n");
    DemoLogf(" w    Write to a Register in the DAA\n");
    DemoLogf(" b    Get Debug data from the DSP\n");
    DemoLogf(" s    Stream receive data to file (toggle)\n");
    DemoLogf(" a    Abort the Burst Receiver\n");
    DemoLogf(" R    Start Burst Receiver. Prompt for file to recv\n");
    DemoLogf(" !    Terminate Watchdog Timer Thread (Crash test)\n");
    DemoLogf(" V    Toggle Burst Sender Verbose mode\n");
    break;

  case 'V':
    {
      if (g_sBurstDevs.mBurstSender.bVerbose) {
        g_sBurstDevs.mBurstSender.bVerbose = OTX_FALSE;
      } else {
        g_sBurstDevs.mBurstSender.bVerbose = OTX_TRUE;
      }
      DemoLogf("Burst Sender Verbose is now %s\n", g_sBurstDevs.mBurstSender.bVerbose ? "ON" : "OFF");
    }
    break;


  case '!':
    {
      DemoLogf("Terminating WatchDog Thread (Crash test)\n");
      g_bWatchDogThreadActive = OTX_FALSE;
    }
    break;


  case 'a':
    {
      if (g_sBurstDevs.mBurstReceiver.pRxFile) {
        DemoLogf("DemoBurstReceiver: Closing file\n");
        fclose(g_sBurstDevs.mBurstReceiver.pRxFile);
        g_sBurstDevs.mBurstReceiver.pRxFile = NULL;
      }
      g_sBurstDevs.mBurstReceiver.bActive = OTX_FALSE;
    }
    break;

  case 'R':
    {
      char sqzBuf[OTX_MAX_PATH];
      static OTX_INT32 nTsToRecord = 0;
      OTX_INT32 i;

      if (DemoIsBurstDeviceEnabled()) {
        if (g_sBurstDevs.mBurstReceiver.bActive == OTX_FALSE || g_sBurstDevs.mBurstReceiver.pRxFile == NULL) {

          for (i = 0; i < DEMO_LI_MAX; i++) {
            /*
            // Connect the Echo Cancel Device
            DemoLogf("Connecting EC device\n");
            result = OtxDrvConnectLogicalDevice(g_hPipes.hDaaRx[i], -1, g_hLogicDevs.hEchoCancel[i], OTX_CST_ECHO_CANCEL_NEAR_END_PIPE); 
            DemoCheckRc(g_hPipes.hDaaRx[i], result);
            result = OtxDrvConnectLogicalDevice(g_hPipes.hBurstTx[i], -1, g_hLogicDevs.hEchoCancel[i], OTX_CST_ECHO_CANCEL_FAR_END_PIPE); 
            DemoCheckRc(g_hPipes.hBurstTx[i], result);
            result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hEchoCancel[i], -1, g_hPipes.hBurstRx[i], -1); 
            DemoCheckRc(g_hLogicDevs.hEchoCancel[i], result);
            
            
            // Initialize the Echo Cancel Device
            DemoLogf("Initalizing EC device\n");
            result = OtxCstEchoCancelInit(
              g_hLogicDevs.hEchoCancel[i], 
              g_demoEnv.law, 
              OTX_CST_ECHO_CANCEL_32_MS); 
            DemoCheckRc(g_hLogicDevs.hEchoCancel[i], result);
            */
            // Enable the Echo Cancel Device
            DemoLogf("Enabling EC device\n");
            result = OtxDrvEnable(g_hLogicDevs.hEchoCancel[i],  OTX_TASK_SYNC, 5000L); 
            DemoCheckRc(g_hLogicDevs.hEchoCancel[i], result);
          }


          /*
          DemoLogf("BurstReceiver. Please enter file to receive: ");
          OsaConsoleGetStr(sqzBuf);
          if (sqzBuf[0] == '\0') {
            break;
          }
          */
          DemoLogf("Burst Receiver. File recording started. Press 'a' to stop recording.\n");
          DemoLogf("Then see files \"%s\", \"%s\", and \"%s\"\n", DEMO_EC_OUTPUT_REC_FILE, DEMO_EC_NE_REC_FILE, DEMO_EC_FE_REC_FILE);
          
          strcpy(sqzBuf, DEMO_EC_OUTPUT_REC_FILE);
          if ((g_sBurstDevs.mBurstReceiver.pRxFile = fopen(sqzBuf, "wb")) == NULL) {
            DemoLogf("Error: Could not open file %s \n", sqzBuf);
            break;
          }
          strcpy(sqzBuf, DEMO_EC_NE_REC_FILE);
          if ((g_sBurstDevs.mBurstReceiver.pRxFile2 = fopen(sqzBuf, "wb")) == NULL) {
            DemoLogf("Error: Could not open file %s \n", sqzBuf);
            break;
          }
          strcpy(sqzBuf, DEMO_EC_FE_REC_FILE);
          if ((g_sBurstDevs.mBurstReceiver.pRxFile3 = fopen(sqzBuf, "wb")) == NULL) {
            DemoLogf("Error: Could not open file %s \n", sqzBuf);
            break;
          }
          g_sBurstDevs.mBurstReceiver.nTsToRecord = nTsToRecord;
          
          g_sBurstDevs.mBurstReceiver.bActive = OTX_TRUE;
        } else {
          DemoLogf("Burst Receiver thread already running.\n");
        }
      } else {
        DemoLogf("The Burst device is not enabled. Command ignored\n");
      }
    }
    break;

  case 'b':
    {
      DemoLogf("Debug Dump for Dsp#%d\n", nCurrentLi);
      OtxDrvDbgDump(g_hPhysDevs.hDsp[nCurrentLi], 0);
    }
    break;


  case 'f':
    // Set Active Pots LiOff-Hook ...
    DemoLogf("Pots Li#%d Off-hook\n", nCurrentLi);
    result = OtxCstPotsOffHook(g_hLogicDevs.hPotsLi[nCurrentLi], 0x1000, 500); 
    if (result == OTX_S_OK) {
      DemoLogf("Check for success/failure in driver DPRINTF.\n");
    } else {
      DemoLogf("Off-hook failure.\n");
    }
    //DemoCheckRc(g_hLogicDevs.hPotsLi[nCurrentLi], result);
    //DemoLogf("Done.\n");
    break;


  case 'n':
    // Set Active Pots LiOn-Hook ...
    DemoLogf("Pots Li#%d On-hook\n", nCurrentLi);
    result = OtxCstPotsOnHook(g_hLogicDevs.hPotsLi[nCurrentLi], 0x1000, 500); 
    if (result == OTX_S_OK) {
      DemoLogf("Check for success/failure in driver DPRINTF.\n");
    } else {
      DemoLogf("On-hook failure.\n");
    }
    //DemoCheckRc(g_hLogicDevs.hPotsLi[nCurrentLi], result);
    //DemoLogf("Done.\n");
    break;

  case 'l':
    // Flash Hook
    DemoLogf("Pots Li#%d Hook Flash\n", nCurrentLi);
    result = OtxCstPotsHookFlash(g_hLogicDevs.hPotsLi[nCurrentLi], 200, 0x1000, 500); DemoCheckRc(g_hLogicDevs.hPotsLi[nCurrentLi], result);
    DemoLogf("Done.\n");
    break;

  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
    // Select current Li
    {
      OTX_INT32 nLi = commandChar - '0';
      
      nCurrentLi = nLi;
      DemoLogf("Li#%d Selected ....\n", nCurrentLi);
    }
    break;

  case 's':
    {
      char sqzBuf[OTX_MAX_PATH];
      
      if (g_fStreamFile) {
        fclose(g_fStreamFile);
        g_fStreamFile = NULL;
        DemoLogf("Closed streaming file\n");
      } else {
        
        DemoLogf("Please enter file to receive: ");
        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0] == '\0') {
          break;
        }
        
        if ((g_fStreamFile = fopen(sqzBuf, "wb")) == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzBuf);
          break;
        }
      }
    }
    break;



  case 'r':
    {
      OTX_UINT32 nCmd = 1; // 1 = Read
      static OTX_UINT8 nRegister = 2; 
      static OTX_UINT8 nDummy2 = 0; 
      static OTX_UINT8 nDummy3 = 0; 
      char strBuf[64];
      
      DemoLogf("Read DAA register. Register [%d]\n ", nRegister);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nRegister = (OTX_UINT8)strtol(strBuf, NULL, 0);
      }
/*      
      DemoLogf("Please enter highway [0x%.2X] ", nHighway);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nHighway = (OTX_UINT8)strtol(strBuf, NULL, 0);
      }
      
      DemoLogf("Please enter TS [0x%.2X] ", nTs);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nTs = (OTX_UINT8)strtol(strBuf, NULL, 0);
      }
*/      
      nCmd = (nDummy3 << 3*8) | (nDummy2 << 2*8) | (nRegister << 1*8) | nCmd;
      result = OtxDrvDbgDump(g_hLogicDevs.hPotsLi[nCurrentLi], nCmd);
      
    }
    break;

  
  case 'w':
    {
      OTX_UINT32 nCmd = 2; // 2 = Write
      static OTX_UINT8 nRegister = 2; 
      static OTX_UINT8 nVal = 0; 
      static OTX_UINT8 nDummy3 = 0; 
      char strBuf[64];
      
      DemoLogf("Write DAA register. Register [%d]\n ", nRegister);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nRegister = (OTX_UINT8)strtol(strBuf, NULL, 0);
      }

      DemoLogf("Please enter value [0x%.2X] ", nVal);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nVal = (OTX_UINT8)strtol(strBuf, NULL, 0);
      }
/*      
      
      DemoLogf("Please enter TS [0x%.2X] ", nTs);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nTs = (OTX_UINT8)strtol(strBuf, NULL, 0);
      }
*/      
      nCmd = (nDummy3 << 3*8) | (nVal << 2*8) | (nRegister << 1*8) | nCmd;
      result = OtxDrvDbgDump(g_hLogicDevs.hPotsLi[nCurrentLi], nCmd);
      
    }
    break;

  


  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;
      
      DemoLogf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);
      
      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoLogf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
      
    }
    break;
    
  case 'q':
  case 'Q':
    // Quit ...
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}



/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->bExitOnError      = OTX_FALSE;
  envP->logFileName       = DEMO_DEFAULT_LOG_FILE;
  envP->outFileName       = NULL;
  envP->law               = OTX_DATA_FORMAT_COMPAND8_U_LAW;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->fxoNo             = DEMO_DEFAULT_FXO_NO;
  envP->nSetId            = OTX_CST_POTS_CFG_SET_USA;
  envP->nBoardType        = OTX_DEVICE_ARNI_6X6_PCI;

  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          envP->logFileName = argv[i]+2;

        } else if (*(argv[i]+1) == 'O') {   // DSP out file
          envP->outFileName = argv[i]+2;

        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);

        } else if (*(argv[i]+1) == 'F') {   // FXO number
          envP->fxoNo = strtoul(argv[i]+2, NULL, 0);

        } else if (OsaToUpper(*(argv[i]+1)) == 'T') {   // Board Type
          if        (OsaToUpper(*(argv[i] + 2)) == 'A') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6X6_PCI_PLUS;
          } else if (OsaToUpper(*(argv[i] + 2)) == 'B') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6_FXO_PCI_PLUS;
          } else if (OsaToUpper(*(argv[i] + 2)) == 'D') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6X3_PCI_PLUS;
          } else if (OsaToUpper(*(argv[i] + 2)) == 'J') {
            envP->nBoardType  = OTX_DEVICE_ARNI_6X6_PCI;
          } else {            
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'C') {   // Companding Law
          if (*(argv[i]+2) == 'A') {   // A-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_A_LAW;
          } else if (*(argv[i]+2) == 'U') {   // U-law
            envP->law = OTX_DATA_FORMAT_COMPAND8_U_LAW;
          } else {
            DemoLogf("Error: Invalid companding law command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }

}


static void DemoHelpPanel()
/*******************************************************************************
  @func
    Displays a help for the command line.
*******************************************************************************/
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName>    : Log file name\n");
  DemoLogf(" /B<boardNo>     : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /F<fxoNo>       : FXO number to configure (Default is %d)\n", DEMO_DEFAULT_FXO_NO);
  DemoLogf(" /T[A|B|C|D|J]   : Board Type used for the demo, where\n");
  DemoLogf("                    A  = Use Arni-6-PCI-Plus Board\n");
  DemoLogf("                    B  = Use Arni-6-FXO-PCI-Plus Board\n");
  DemoLogf("                    D  = Use Arni-6x3-PCI-Plus Board\n");
  DemoLogf("                    J  = Use Arni-6-PCI Board\n");
  DemoLogf(" /C[A|U]         : Voice Compression Codec, where\n");
  DemoLogf("                    A  = A-law\n");
  DemoLogf("                    U  = u-law (default)\n");
  DemoLogf(" /O<fileName>    : DSP Out file name (for debugging only)\n");
}


static void DemoLogf(char *formatStr, ...)
/*******************************************************************************
  @func
    Displays a message on the screen and prints it to the logfile.
*******************************************************************************/
{
  va_list argList;

  OsaMutexLock(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  OsaMutexUnlock(&g_demoLogfCritSec);
}



/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  //OTX_INT32  i;
  OTX_INT32  exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char       commandChar;

	OsaConsoleInit();

  OsaMutexInit(&g_demoLogfCritSec);

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    exit(-1);
  }

  DemoLogf("%s, %s, Odin TeleSystems Inc.\n", argv[0], __DATE__);
  DemoLogf("Arni-6x6-PCI, Basic Demo Program.\n");

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);
  DemoLogf("FXO   number %d\n", g_demoEnv.fxoNo);
  if (g_demoEnv.outFileName) {
    DemoLogf("Out file to be loaded is %s\n", g_demoEnv.outFileName);
  }


  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    DemoExit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    DemoExit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }


  // Create Event Queues for Driver --> Application Communications
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Arni-16-PCI board
  DemoOpenPhysicalDevices();

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  {
    // Read CPLD Build Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(g_hPhysDevs.hBoardController0, result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);

    free(pszCpldBuildNumbers);

  }

  // Enable Physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnablePhysicalDevices();

  OsaThreadStart(DemoIdleFunction);

  // Create and get handles to the Logical devices needed by this application
  DemoCreateLogicalDevices();

  // Create and get handles to the Pipe devices needed by this application
  // Connect Pipes to Physical and Logical Devices
  //DemoLogf("Skipping pipe connections for now\n");
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoInitializeLogicalDevices();

  // Enable Logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnableLogicalDevices();

#if 0
  // Enable the Watchdog Timer itself
  result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_ENABLED,  OTX_TRUE); DemoCheckRc(g_hPhysDevs.hBoardController0, result);

  // Start the Watchdog timer Reset Thread, which will keep the watchdog alive
  OsaThreadStart(DemoWatchDogFunction);
#endif
  
  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....

  DemoLogf("\nPress 'h' for help, 'q' to quit ...\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
		exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Delete allocated resources
  DemoExit(0);

  return 0;
}

/*******************************************************************************
*******************************************************************************/
static int DemoExit(int nExitCode)
{

  DemoLogf("\nExiting...\n");

  // Important to disable the Watchdog Timer since we will not be reseting it after we exit
  // the application. 
  OtxDrvSetAttributeValueBool(g_hPhysDevs.hBoardController0, OTX_ATTR_BRD_WATCHDOG_TIMER_ENABLED,  OTX_FALSE);

  g_bWatchDogThreadActive = OTX_FALSE;

  if (DemoIsBurstDeviceEnabled()) {
    if (g_sBurstDevs.mBurstReceiver.pRxFile) {
      DemoLogf("DemoBurstReceiver: Closing file\n");
      fclose(g_sBurstDevs.mBurstReceiver.pRxFile);
      g_sBurstDevs.mBurstReceiver.pRxFile = NULL;
    }
    if (g_sBurstDevs.mBurstReceiver.pRxFile2) {
      DemoLogf("DemoBurstReceiver: Closing file\n");
      fclose(g_sBurstDevs.mBurstReceiver.pRxFile2);
      g_sBurstDevs.mBurstReceiver.pRxFile2 = NULL;
    }
    if (g_sBurstDevs.mBurstReceiver.pRxFile3) {
      DemoLogf("DemoBurstReceiver: Closing file\n");
      fclose(g_sBurstDevs.mBurstReceiver.pRxFile3);
      g_sBurstDevs.mBurstReceiver.pRxFile3 = NULL;
    }
    g_sBurstDevs.mBurstReceiver.bActive = OTX_FALSE;
    
    if (g_sBurstDevs.mBurstSender.pTxFile) {
      DemoLogf("mBurstSender: Closing file\n");
      fclose(g_sBurstDevs.mBurstSender.pTxFile);
      g_sBurstDevs.mBurstSender.pTxFile = NULL;
    }
#ifdef DEMO_EC_DEBUG
    if (g_sBurstDevs.mBurstSender.pTxFile2) {
      DemoLogf("mBurstSender: Closing file\n");
      fclose(g_sBurstDevs.mBurstSender.pTxFile2);
      g_sBurstDevs.mBurstSender.pTxFile2 = NULL;
    }
#endif
    g_sBurstDevs.mBurstSender.bActive   = OTX_FALSE;
  }

  OsaThreadSleep(100);   // give threads a chance to exit

  // Clean up
  OsaConsoleExit();
  OtxDrvReleaseLib();

  exit(nExitCode);
}
